from scipy import stats

class Initializer(object):
    """Base class for weight initializers."""
    def __init__(self):
        pass

    def normal(self, shape, mean=0.0, std=1.0):
        return stats.truncnorm.rvs(-2, 2, loc=mean, scale=std, size=shape)

    def uniform(self, shape, low=0.0, high=1.0):
        return stats.uniform.rvs(loc=low, scale=high-low, size=shape)

    def xavier(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0/shape[0]**0.5, size=shape)

    def he(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=2.0/shape[0]**0.5, size=shape)

    def lecun(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0/shape[0]**0.5, size=shape)

    def bengio(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0/shape[0]**0.5, size=shape)

    def glorot(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0/shape[0]**0.5, size=shape)

    def he_uniform(self, shape):
        return stats.uniform.rvs(scale=6.0/shape[0]**0.5, size=shape)

    def lecun_uniform(self, shape):
        return stats.uniform.rvs(scale=1.0/shape[0]**0.5, size=shape)

    def bengio_uniform(self, shape):
        return stats.uniform.rvs(scale=1.0/shape[0]**0.5, size=shape)

    def glorot_uniform(self, shape):
        return stats.uniform.rvs(scale=1.0/shape[0]**0.5, size=shape)

    def he_normal(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=2.0/shape[0]**0.5, size=shape)

    def lecun_normal(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0/shape[0]**0.5, size=shape)

    def bengio_normal(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0/shape[0]**0.5, size=shape)

    def glorot_normal(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0/shape[0]**0.5, size=shape)

    def identity(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0, size=shape)

    def zero(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=0.0, size=shape)

    def one(self, shape):
        return stats.truncnorm.rvs(-2, 2, scale=1.0, size=shape)

    def constant(self, shape, value=0.0):
        return stats.truncnorm.rvs(-2, 2, scale=value, size=shape)

    def random(self, shape):
        return stats.truncnorm.rvs(-2, 2, size=shape)

    def uniform_random(self, shape, low=0.0, high=1.0):
        return stats.uniform.rvs(loc=low, scale=high-low, size=shape)

    def normal_random(self, shape, mean=0.0, std=1.0):
        return stats.truncnorm.rvs(-2, 2, loc=mean, scale=std, size=shape)

    def constant_random(self, shape, value=0.0):
        return stats.truncnorm.rvs(-2, 2, scale=value, size=shape)

    def random_random(self, shape):
        return stats.truncnorm.rvs(-2, 2, size=shape)

    def truncated_normal(self, shape, mean=0.0, std=1.0, low=-2.0, high=2.0):
        return stats.truncnorm.rvs(low, high, loc=mean, scale=std, size=shape)

    def truncated_uniform(self, shape, low=-1.0, high=1.0):
        return stats.uniform.rvs(loc=low, scale=high-low, size=shape)

    def truncated_normal_random(self, shape, mean=0.0, std=1.0, low=-2.0, high=2.0):
        return stats.truncnorm.rvs(low, high, loc=mean, scale=std, size=shape)

    def truncated_uniform_random(self, shape, low=-1.0, high=1.0):
        return stats.uniform.rvs(loc=low, scale=high-low, size=shape)

    def truncated_normal_constant(self, shape, mean=0.0, std=1.0, low=-2.0, high=2.0, value=0.0):
        return stats.truncnorm.rvs(low, high, loc=mean, scale=std, size=shape) + value

    def truncated_uniform_constant(self, shape, low=-1.0, high=1.0, value=0.0):
        return stats.uniform.rvs(loc=low, scale=high-low, size=shape) + value

    def truncated_normal_random_constant(self, shape, mean=0.0, std=1.0, low=-2.0, high=2.0, value=0.0):
        return stats.truncnorm.rvs(low, high, loc=mean, scale=std, size=shape) + value

    def truncated_uniform_random_constant(self, shape, low=-1.0, high=1.0, value=0.0):
        return stats.uniform.rvs(loc=low, scale=high-low, size=shape) + value

    def truncated_normal_constant_random(self, shape, mean=0.0, std=1.0, low=-2.0, high=2.0, value=0.0):
        return stats.truncnorm.rvs(low, high, loc=mean, scale=std, size=shape) + value

    def truncated_uniform_constant_random(self, shape, low=-1.0, high=1.0, value=0.0):
        return stats.uniform.rvs(loc=low, scale=high-low, size=shape) + value

    def truncated_normal_constant_random_constant(self, shape, mean=0.0, std=1.0, low=-2.0, high=2.0, value=0.0):
        return stats.truncnorm.rvs(low, high, loc=mean, scale=std, size=shape) + value

    def truncated_uniform_constant_random_constant(self, shape, low=-1.0, high=1.0, value=0.0):
        return stats.uniform.rvs(loc=low, scale=high-low, size=shape) + value

    def truncated_normal_constant_random_constant_random(self, shape, mean=0.0, std=1.0, low=-2.0, high=2.0, value=0.0):
        return stats.truncnorm.rvs(low, high, loc=mean, scale=std, size=shape) + value



from nillanet.model import NN
from nillanet.activations import Activations
from nillanet.loss import Loss
from nillanet.distributions import Distributions

d = Distributions()
x,y = d.arithmetic_distribution(10,"subtract")
print(x.shape)
print(y.shape)

a = Activations()
activation = a.tanh
derivative1 = a.tanh_derivative
resolver = a.linear
derivative2 = a.linear_derivative

l = Loss()
loss = l.mse
derivative3 = l.mse_derivative

input = x
output = y
features = x.shape[1]
architecture = [2,4,1]
learning_rate = 0.01
epochs = 1000

model = NN(features,architecture,activation,derivative1,resolver,derivative2,loss,derivative3,learning_rate)
model.train(input,output,epochs,verbose=True,step=100,autosave=True)
prediction = model.predict(input)

print("prediction")
print(prediction)
print("expected")
print(y)

from nillanet.io import IO
io = IO()
best = io.load(model.backup)
prediction = best.predict(x)
print("best")
print(prediction)


import cupy as cp
import pickle

class IO(object):
  """Helper functions for NN class.
  """

  def __init__(self):
    pass

  def save(self, model, filename):
    """Serialize the model to disk using pickle.

       Args:
           filename: Path to the output pickle file.
            
       CuPy arrays are picklable
    """ 

    with open(filename, "wb") as fh:
      pickle.dump(model, fh)

  def load(self,filename):
    """Read serialized file
    
    Args:
        filename: Path to the model pickle file.

    """
    with open(filename, "rb") as r:
      model = pickle.load(r)
    return model




from nn.model import NN
from nn.tools import Tools
from tests.distributions import Distributions

distribution = Distributions()
x,y = distribution.tally(10,4,"summation")
print(x.shape)
print(y.shape)
print(x)
print(y)

tools = Tools()
input = x
output = y
architecture = [2,4,1]
activation = tools.sigmoid
derivative1 = tools.sigmoid_derivative
classifier = tools.linear
derivative2 = tools.linear_derivative
loss = tools.mse
derivative3 = tools.mse_derivative
learning_rate = 0.1

model = NN(input,output,architecture,activation,derivative1,classifier,derivative2,loss,derivative3,learning_rate)
model.train(10000,1)
prediction = model.predict(x)

print("prediction")
print(prediction)
print("expected")
print(y)



from scipy import stats
import sys

class Initializer(object):
    """Weight distributions for custom initializations."""
    def __init__(self, distribution=None, low=0.0, high=1.0, mean=0.0, std=1.0):
        self.distribution = distribution
        if distribution is None:
            self.distribution = self.normal
        self.low = low
        self.high = high
        self.mean = mean
        self.std = std

    def __call__(self, shape):
        return self.distribution(self, shape)

    def normal(self, shape):
        """bell shaped distribution"""
        return stats.truncnorm.rvs(self.low, self.high, loc=self.mean, scale=self.std, size=shape)

    def uniform(self, shape):
        """plateau shaped distribution"""
        return stats.uniform.rvs(loc=self.low, scale=self.high-self.low, size=shape)

    def xavier(self, shape):
        """for tanh or sigmoid activation function"""
        return stats.truncnorm.rvs(self.low, self.high, scale=1.0/shape[0]**0.5, size=shape)

    def he(self, shape):
        """for relu activation function"""
        return stats.truncnorm.rvs(self.low, self.high, scale=2.0/shape[0]**0.5, size=shape)

if __name__ == '__main__':
    shape = (3, 3)
    low,high,mean,std = (
        0.0, 1.0, 0.0, 1.0
    )
    if len(sys.argv) >= 3:
        low = sys.argv[1]
        high = sys.argv[2]
    if len(sys.argv) >= 5:
        mean = sys.argv[3]
        std = sys.argv[4]
    print("shape %s low %s high %s mean %s std %s" % (shape, low, high, mean, std))
    print("normal")
    intl = Initializer(distribution=Initializer.normal, low=low, high=high, mean=mean, std=std)
    print(intl.normal(shape))
    print("uniform")
    intl = Initializer(distribution=Initializer.uniform, low=low, high=high, mean=mean, std=std)
    print(intl.uniform(shape))
    print("xavier")
    intl = Initializer(distribution=Initializer.xavier, low=low, high=high, mean=mean, std=std)
    print(intl.xavier(shape))
    print("he")
    intl = Initializer(distribution=Initializer.he, low=low, high=high, mean=mean, std=std)
    print(intl.he(shape))
